package board;

import game.Coordinate;
import player.Player;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

/**
 * Created by danielford on 11/02/16.
 */
public class Board implements Cloneable {

    public Player[][] board;

    // The most recently taken move
    private Coordinate lastMove;
    private int winningScore;
    private int dimensions;
    private int counters = 0;
    private int maxCounters;
    private List<Player> players;
    private Player currentPlayer;
    private Player firstPlayer;
    private int currentPlayerIndex;
    private static final Logger LOGGER = Logger.getLogger(Board.class.getName());

    public Board(int dimensions, int winningScore) {

        this.dimensions = dimensions;
        this.winningScore = winningScore;
        this.maxCounters = dimensions * dimensions;

        LOGGER.info("Created board, dimensions: " + dimensions + " maxScore: " + winningScore);

        // Creates a 19 x 19 board
        board = new Player[this.dimensions][this.dimensions];
        players = new ArrayList<>();

    }

    /**
     * @return Board, Returns the board
     */
    public Player[][] getBoard() {
        return board;
    }

    /**
     * Returns the actual size of the board
     * @return Integer, the dimensions.
     */
    public int getDimensions() { return dimensions; }

    /**
     * Creates a move for a player at the coords and increases counters used
     * @param coordinate, The coordinate to try placing in
     * @param player, The player to place at the coordinate
     * @return Board, Returns the data representation of the board
     */
    public Player[][] createMove(Coordinate coordinate, Player player) {

        if(coordinate == null) throw new Error("Coordinates missing");

        if ( !isLegal( coordinate ) ) throw new Error("Space is already taken");

        counters = counters + 1;

        lastMove = coordinate;

        board[coordinate.row][coordinate.column] = player;

        return board;

    }

    /**
     * If we are not out of bounds, return the player in the current board position
     * @param coordinate, The coordinate to check
     * @return Player, The player in the coordinates or null
     */
    public Player getPosition(Coordinate coordinate) {

        if ( coordinate != null ) {
            return coordinate.row == -1 || coordinate.row > dimensions -1 || coordinate.column == -1 || coordinate.column > dimensions -1 ? null : board[coordinate.row][coordinate.column];
        } else {
            return null;
        }

    }


    public void resetMove(Coordinate coordinate) {
        counters -= 1;
        board[coordinate.row][coordinate.column] = null;
    }

    /**
     * A way of checking the Rows and/or columns from a starting counter to see if there is a winner
     * @param checkRow {Boolean}, Should we check the rows for a winner
     * @param checkCol {Boolean}, Should we check the columns for a winner
     * @return {Boolean}, True if a player has won
     */
    private boolean findWinner(Boolean checkRow, Boolean checkCol, Boolean checkDiag) {

        if ( lastMove == null ) return false;

        boolean notFinished = true;
        Coordinate pointer = lastMove;
        int newCol = pointer.column;
        int newRow = pointer.row;
        int currentScore = 0;

        while ( currentScore != winningScore ) {

            // We are moving the column value i.e. checkCol
            if ( checkCol && !checkRow ) {
                newCol = notFinished ? pointer.column - 1 : pointer.column + 1;
            } else if ( checkRow && !checkCol) {
                // We are moving the row value i.e.  checkRow
                newRow = notFinished ? pointer.row - 1 : pointer.row + 1;
            } else if ( checkDiag ) {
                newCol = notFinished ? pointer.column - 1 : pointer.column + 1;
                newRow = notFinished ? pointer.row - 1 : pointer.row + 1;
            } else {
                newCol = notFinished ? pointer.column + 1 : pointer.column - 1;
                newRow = notFinished ? pointer.row - 1 : pointer.row + 1;
            }

            // If there is a player in the position, and they are the current player, increase counter
            if ( getPosition(pointer) != null &&
                    getPosition(pointer).getPlayerIdentity() == getPosition(lastMove).getPlayer().getPlayerIdentity() ) {

                pointer = new Coordinate(newCol, newRow);
                currentScore++;

            } else if ( notFinished ) {

                // Else start counting the counters on the other side of placed counter
                notFinished = false;
                int resetCol = lastMove.column;
                int resetRow = lastMove.row;

                if ( checkCol && !checkRow ) {
                    resetCol = resetCol + 1;
                } else if ( checkRow && !checkCol ) {
                    resetRow = resetRow + 1;
                } else if ( checkDiag ) {
                    resetCol = resetCol + 1;
                    resetRow = resetRow + 1;
                } else {
                    resetCol = resetCol - 1;
                    resetRow = resetRow + 1;
                }

                pointer = new Coordinate(resetCol, resetRow);

            } else {

                // We are finished and currentScore has not reached winningScore
                return false;

            }

        }

        // We've reached the winning Score
        return true;

    }

    /**
     * Checks the columns in a row to check for a winner
     * @return {Boolean}, true if a player has won
     */
    private boolean checkColumns() {
        return findWinner(false, true, false);
    }

    /** Checks the rows in a column to check for a winner
     *  @return {Boolean}, true if a player has won
     */
    private boolean checkRows() {
        return findWinner(true, false, false);
    }

    /**
     * Checks the diagonal combinations for a winner
     * @return {Boolean}, true if a player has won
     */
    private boolean checkRightDiagonals() {
        return findWinner(true, true, true);
    }

    private boolean checkLeftDiagonals() {
        return findWinner(true, true, false);
    }

    public void createPlayer(Player player) {
        players.add(player);
    }

    public Player getPlayer(int index) {
        return players.get(index);
    }

    public boolean isLegal( Coordinate coordinate ) { return board[coordinate.row][coordinate.column] == null; }

    public WinState checkWinner() {

        Player currentPlayer = getPosition(lastMove);

        if ( checkRows() || checkColumns() || checkRightDiagonals() || checkLeftDiagonals() ) {
            // One of the players have won
            return new WinState(currentPlayer, true);
        } else if ( counters >= maxCounters ) {
            // The players have drawn
            return new WinState(null, false);
        } else {
            // No winner and we can still go due to maxCounters not reached
            return new WinState();
        }

    }

    public Player setCurrentPlayer(int index) {
        currentPlayer = players.get(index);
        currentPlayerIndex = index;
        return currentPlayer;
    }

    /**
     * This returns all possible moves, i.e. all places on the board that are null.
     * @return List legalMoves
     */
    public List<Coordinate> getLegalMoves() {

        List<Coordinate> legalMoves = new ArrayList<Coordinate>();

        for ( int i = 0; i < dimensions; i++ ) {

            for ( int j = 0; j < dimensions; j++ ) {

                Coordinate attempt = new Coordinate(i, j);

                if ( isLegal( attempt ) ) {

                    legalMoves.add( attempt );

                }

            }

        }

        return legalMoves;

    }

    /**
     * Clones the necessary elements in the board so that the Objects are not
     * tied together and a change to one will not affect the other.
     * @return Board, the newly cloned board
     */
    public Board clone () {

        Board newBoard = null;

        try {

            newBoard = (Board) super.clone();
            newBoard.board = cloneBoard();

        } catch (CloneNotSupportedException e) { e.printStackTrace(); }


        return newBoard;

    }

    // Clones the board by reading each value from the old board into the new board.
    private Player[][] cloneBoard() {

        Player[][] newBoard = new Player[dimensions][dimensions];

        for ( int i = 0; i < dimensions; i++) {
            for ( int j = 0; j < dimensions; j++) {
                newBoard[i][j] = board[i][j];
            }
        }

        return newBoard;

    }

    public Player getCurrentPlayer() { return currentPlayer; }

    public Player setFirstPlayer(Player player) {
        return firstPlayer = player;
    }

    public Player getFirstPlayer() {
        return firstPlayer;
    }

    public int getCounters() { return counters; }

    public int getCurrentPlayerIndex() { return currentPlayerIndex; }

    public Coordinate getLastMove() { return lastMove; }

    /**
     * Useful for debugging, will print out the current board state where player 1 is X
     * and player 2 is O.
     */
    public void visualiseBoard(){

        System.out.println();
        for(int row = 0; row < dimensions; row++){
            for(int col = 0; col < dimensions; col++){

                System.out.print("|");

                if(board[row][col] == players.get(0)){
                    System.out.print(" X ");
                } else if( board[row][col] == players.get(1) ){
                    System.out.print(" O ");
                } else {
                    System.out.print("   ");
                }

                if ( col == dimensions - 1 ) {
                    System.out.println("|");
                }

            }
            System.out.println();
        }

    }

    
}