package game;

/**
 * Created by danielford on 11/02/16.
 */

import javafx.animation.FadeTransition;
import javafx.geometry.HPos;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.layout.*;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Font;
import javafx.scene.text.FontWeight;
import javafx.scene.text.Text;
import javafx.stage.*;
import javafx.util.Duration;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import static javafx.scene.layout.GridPane.setMargin;


public class Game {

    private int boardSize = 600;
    private Stage stage;
    private GridPane boardUI;
    private GridPane sidePanel;
    private PlayerTracker playerTracker;
    private StackPane stack;
    private Counter[][] counterBoard;
    private int dimensions;

    /**
     * @param gameStage, the primary stage from Application
     * @param dimensions, the board dimensions
     *
     */
    public Game(Stage gameStage, int dimensions) {

        stage = gameStage;

        this.dimensions = dimensions;

        setupStage(stage);

    }

    /**
     * The main setting up of the UI
     * @param stage, the primary stage from Application
     * @return {Object}, the set up UI
     */
    private Stage setupStage(Stage stage) {

        stack = new StackPane();
        stack.getStylesheets().add("resources/styles.css");

        stack.getChildren().add(mainScene());

        stage.setTitle("Gomoku - Playing Game");

        stage.setScene(new Scene(stack));
        stage.setResizable(false);
        stage.sizeToScene();
        stage.show();

        return stage;
    }

    private BorderPane mainScene() {

        BorderPane border = new BorderPane();

        border.setCenter(createBoard());
        border.setRight(createSidePanel());

        return border;

    }

    private StackPane createOverlay() {

        StackPane stackPane = new StackPane();
        Rectangle rectangle = new Rectangle();

        stackPane.getStyleClass().add("overlay-container");
        rectangle.getStyleClass().add("overlay");

        rectangle.setHeight(stage.getHeight());
        rectangle.setWidth(stage.getWidth());

        stackPane.getChildren().add(rectangle);

        return stackPane;

    }


    /**
     * Creates the UI representation of the board.
     * @return {Object}, the Object which represents the UI board
     */
    private Pane createBoard() {

        boardUI = createContainer();
        boardUI.setPadding(new Insets(20, 40, 20, 20));
        boardUI.setPrefSize(600, 600);

        int size = dimensions;
        int tileSize = boardSize / size;
        counterBoard = new Counter[ size ][ size ];

        List<Tile> tileUI = new ArrayList<>();
        List<Intersection> intersections = new ArrayList<>();
        List<Counter> counters = new ArrayList<>();

        for(int i = 0; i < size; i++) {
            for(int j = 0; j < size; j++) {

                tileUI.add( new Tile(tileSize, i, j, dimensions) );
                intersections.add( new Intersection(tileSize, i, j));
                counters.add( new Counter(tileSize, i, j) );

                counterBoard[ j ][ i ] = counters.get( counters.size() - 1 );

            }
        }

        boardUI.getChildren().addAll( tileUI );
        boardUI.getChildren().addAll( intersections );
        boardUI.getChildren().addAll( counters );

        return boardUI;

    }

   public Counter getTile(Coordinate coordinate) {
       return counterBoard[coordinate.row][coordinate.column];
    }

    private Pane createSidePanel() {

        BorderPane pane = new BorderPane();
        pane.getStyleClass().addAll("container", "side-panel");

        sidePanel = new GridPane();
        pane.setPrefWidth(150);
        pane.setPadding(new Insets(20, 20, 20, 20));

        playerTracker = new PlayerTracker();

        sidePanel.add(playerTracker, 0, 0);

        pane.setTop( sidePanel );

        Button button = new Button();
        button.getStyleClass().addAll("btn", "btn-warning");
        button.setText("End Game");

        button.setTranslateX(25);

        button.setOnMouseClicked(event -> {
            resetGame();
        });

        pane.setBottom( button );

        return pane;

    }

    private GridPane createContainer() {

        GridPane pane = new GridPane();
        pane.getStyleClass().add("container");

        return pane;

    }

    /**
     * A publicly accessible method of retrieving the board.
     * @return {Object}, returns the Boards UI
     */
    public GridPane getBoardUI() {
        return boardUI;
    }

    public GridPane getSidePanel() { return sidePanel; }

    public PlayerTracker getPlayerTracker() { return playerTracker; }

    public Stage getStage() {
        return stage;
    }

    public void setWinnerOverlay(String winner) {

        // The components to add to the Stackpane stack
        StackPane overlay = createOverlay();
        GridPane contents = new GridPane();

        // The components to add to the Pane contents
        Text winnerTitle = new Text();
        Text result = new Text();
        Button replay = new Button();

        // Create and add the column
        ColumnConstraints col1 = new ColumnConstraints();
        col1.setPercentWidth(100);
        contents.getColumnConstraints().add(col1);

        // Customising and laying out Text winnerTitle
        winnerTitle.setText("Game Finished!");
        winnerTitle.setFill(Paint.valueOf("WHITE"));
        winnerTitle.setFont(Font.font("Arial", FontWeight.BOLD, 60));

        // Customising and laying out Text result
        result.setText(!Objects.equals(winner, "Draw") ? winner + " has won." : "The game was a draw.");
        result.setFill(Paint.valueOf("WHITE"));
        result.setFont(Font.font("Arial", FontWeight.BOLD, 35));
        replay.getStyleClass().addAll("btn", "btn-success");

        // Customising  the replay button
        replay.setText("Rematch?");
        replay.setOnMouseClicked(event -> {
            resetGame();
        });

        // Horizontally align the columns
        GridPane.setHalignment(winnerTitle, HPos.CENTER);
        GridPane.setHalignment(result, HPos.CENTER);
        GridPane.setHalignment(replay, HPos.CENTER);

        // Add the contents to the columns and vertically align them
        contents.add(winnerTitle, 0, 0);
        contents.add(result, 0, 1);
        contents.add(replay, 0, 2);
        contents.setAlignment(Pos.CENTER);

        setMargin(replay, new Insets(20, 0, 0, 0));

        // Add the contents to the overlay and to the main screen.
        overlay.getChildren().add(contents);
        stack.getChildren().add(overlay);

        // Fade in the overlay
        FadeTransition ft = new FadeTransition(Duration.millis(400), overlay);
        ft.setFromValue(0);
        ft.setToValue(1);
        ft.play();

    }

    private void resetGame() {
        stage.close();
        new SetupGame( new Stage() );
    }

}
