package mcts;

import board.Board;
import game.Coordinate;
import javafx.geometry.Pos;
import player.Heuristic;
import player.PositionAndScore;

import javax.swing.text.Position;
import java.util.*;
import java.util.logging.Logger;

/**
 * Created by danielford on 08/04/16.
 */
public class MonteCarlo {

    private long start;
    private final Board board;
    private final int maxMoves;
    private final long timeLimit;
    private final boolean useHeuristic;
    private static final Logger LOGGER = Logger.getLogger(MonteCarlo.class.getName());

    public MonteCarlo(Board board, long timeLimit, int maxMoves, boolean useHeuristic) {
        this.board = board;
        this.maxMoves = maxMoves;
        this.timeLimit = timeLimit;
        this.useHeuristic = useHeuristic;
    }

    public Coordinate makeMove() {


        // If there are no moves, use the heuristic to get the central position
        if ( board.getCounters() == 0 ) return new Heuristic( board, board.getCurrentPlayer() ).makeMove();

        // Create the monte carlo thread
        MonteCarloThread monteThread = new MonteCarloThread( board, maxMoves, useHeuristic );
        Thread t = new Thread(monteThread);
        t.start();

        // Start the counter
        start = System.currentTimeMillis();

        // While the thread is not in the finished state, continue checking if it is time to interrupt
        while ( !monteThread.getFinished() ) {

            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                LOGGER.severe(String.valueOf(e));
            }

            if ( finished() && !monteThread.getInterrupted() ) {
                monteThread.setInterrupted();
            }

        }

        // Get the best moves so far
        List<Node> bestMoves = monteThread.getBestNodes();

        // Sort them by how many visits they have
        bestMoves.sort((Node o1, Node o2) -> Double.compare( o1.getVisits(), o2.getVisits() ) );

        // Condense the moves by counting the duplicates in an attempt of creating 'certainty'
        List<PositionAndScore> scores = scoreVisits( bestMoves );

        // Sort the nodes by the newly most visited node
        scores.sort((PositionAndScore o1, PositionAndScore o2) -> Double.compare( o1.getScore(), o2.getScore() ) );

        // Return the most visited node
        return scores.get( scores.size() - 1 ).getCoordinate();

    }

    private boolean finished() {
        return System.currentTimeMillis() > start + timeLimit;
    }

    private List<PositionAndScore> scoreVisits( List<Node> bestMoves ) {
        List<PositionAndScore> scores = new ArrayList<>();
        List<Coordinate> test = new ArrayList<>();

        for ( Node move : bestMoves ) {

            Coordinate lastMove = move.getBoard().getLastMove();

            if ( test.isEmpty()  ) {
                test.add( new Coordinate( lastMove.column, lastMove.row ));
                scores.add( new PositionAndScore( test.get( test.size() - 1), 1, true ) );
            } else {

                boolean added = false;
                for ( int i = 0; i < test.size(); i++ ) {
                    if ( test.get(i).row == lastMove.row && test.get(i).column == lastMove.column ) {
                        scores.get( i ).setScore(scores.get(i).getScore() + 1);
                        added = true;
                        break;
                    }
                }

                if ( !added ) {
                    test.add( new Coordinate( lastMove.column, lastMove.row ));
                    scores.add( new PositionAndScore( test.get( test.size() - 1), 1, true ) );
                }



            }

        }

        return scores;
    }

}
