package mcts;

import board.Board;
import board.WinState;
import game.Coordinate;
import player.Heuristic;
import player.Player;
import player.PositionAndScore;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Random;
import java.util.logging.Logger;

/**
 * Created by danielford on 27/04/16.
 */
public class MonteCarloThread implements Runnable {

    private static final Logger LOGGER = Logger.getLogger(MonteCarlo.class.getName());
    private Board board;
    private int maxMoves;
    private boolean useHeuristic;
    private List<Node> bestNodes;
    private boolean finished = false;
    private boolean interrupted = false;

    private Board boardState;
    static double epsilon = 1e-6;

    private static Random random = new Random();


    public MonteCarloThread(Board board, int maxMoves, boolean useHeuristic) {
        this.board = board;
        this.maxMoves = maxMoves;
        this.useHeuristic = useHeuristic;
    }

    @Override
    public void run() {

        finished = false;
        interrupted = false;
        bestNodes = new ArrayList<>();

        while ( !interrupted ) {

            // The rootNode has not been expanded and therefore will not yet have any children
            Node rootNode = new Node(board.clone(), null);

            // While time not met, run the MCTS algorithm and count how many games we have done

            // Start the timer for the simulations
            long start = System.currentTimeMillis();
            LOGGER.info("Running simulations");

            mcts(rootNode);

            // End simulations
            LOGGER.info("Finished simulations: " + (System.currentTimeMillis() - start) + "ms");

            // Return the best move from the tree, scored through back propagation
            addBestNode(rootNode.getPlayedChildren());

        }

        finished = true;
        
    }

    /**
     * Calls all methods of the MCTS for a maximum number of moves which acts a bit like depth limiting
     * @param rootNode, the Node representing the current 'real' board state to begin on
     */
    private void mcts( Node rootNode ) {

        for ( int i = 0; i < maxMoves; i ++) {

            Node node = rootNode;
            boardState = rootNode.getBoard().clone();

            // Gathers the child nodes which have not yet been played
            node.findUnplayed();

            // Selection
            // On the initial run, selection will not do anything due to not yet having any statistics for the children
            node = selection( node );

            // Expansion
            // This will expand the current node make it so we can generate statistics for the child moves
            if ( getWinner( node.getBoard() ) == null && !node.getUnplayedChildren().isEmpty() ) {

                Node expandedNode = expansion( node );

                if ( expandedNode != null ) {
                    node = expandedNode;
                }

            }


            // Simulation
            simulation();


            // Back-propagation
            backpropagate( node );

            if ( interrupted ) break;

        }

    }

    /**
     * Selection phase of the MCTS, makes UCT selections of the statistics exist in the children
     * i.e. all children have been played meaning the node is fully expanded.
     * @param node, the current node to begin the selection phase on
     * @return Node, the UCT selected move
     */
    private Node selection( Node node ) {

        while ( node.getUnplayedChildren().isEmpty() && !node.getPlayedChildren().isEmpty() ) {

            node = makeUCTMove( node, node.getPlayedChildren() );

            // Create the move UCT move on the board
            // ( Players will switch due to node being re-assigned to child move )
            boardState = node.getBoard().clone();

        }

        return node;

    }


    /**
     * Expands the current node by playing a random, unplayed move and making it the current Node
     * @param node, the not fully expanded node to expand
     * @return Node, the new current node
     */
    private Node expansion( Node node ) {

        Node move = node.expand();

        boardState = move.getBoard().clone();

        return move;

    }


    /**
     * Simulation phase of the MCTS, makes either random or heuristically selected moves
     * on the board until some end state has been met.
     */
    private void simulation() {

        while ( !hasSimulationMoves() ) {

            // The game has ended i.e. a Draw or a player has won, so exit
            if (  gameEnded(boardState) != null ) {
                break;
            }

            // Switch player
            boardState.setCurrentPlayer( boardState.getCurrentPlayerIndex() == 0 ? 1 : 0 );

            // Create a random move on the board for the current player
            boardState.createMove(
                    useHeuristic ?
                            // Best move from heuristic
                            new Heuristic( boardState, boardState.getCurrentPlayer()).makeMove() :
                            // Random with the heuristic
//                            new Heuristic( boardState, boardState.getCurrentPlayer() ).getMovesWithScores().get( random.nextInt( new Heuristic( boardState, boardState.getCurrentPlayer() ).getMovesWithScores().size() ) ).getCoordinate() :
                            // Random without heuristic
                            boardState.getLegalMoves().get( random.nextInt( boardState.getLegalMoves().size()) ),
                    boardState.getCurrentPlayer()
            );

        }

    }

    /**
     * Backpropagates the results back up through the tree
     * from the perspective of the current winner. The value will only be incremented in the tree
     * if the player at that level matches the winner.
     * @param node, the node to begin backpropagation from.
     */
    private void backpropagate( Node node ) {

        WinState gameEnd = gameEnded( boardState );
        Player winner = getWinner( boardState );

        // If the parent is null, we are back at the parent
        while ( node != null ) {

            node.visitNode(gameEnd, winner != null && winner.equals(node.getPlayer()) );

            node = node.getParent();

        }

    }

    /**
     *
     * @param parent, the parent node
     * @param children, the children to select the appropriate current node from using UCT
     * @return Node, the best node as determined by UCT
     */
    private Node makeUCTMove ( Node parent, List<Node> children ) {

        Node best = null;
        double bestScore = Double.NEGATIVE_INFINITY;

        for ( Node child : children ) {

            double uctValue = ( child.getWins() / child.getVisits() ) + epsilon * Math.sqrt( Math.log( parent.getVisits() /child.getVisits() ) );

            if ( uctValue > bestScore ) {
                best = child;
                bestScore = uctValue;
            }

        }

        return best;

    }

    /**
     * Checks if there are any legal moves for the simulation to play
     * @return Boolean, True if it has legal moves available.
     */
    private boolean hasSimulationMoves() {

        if ( useHeuristic ) {
            return new Heuristic( boardState, boardState.getCurrentPlayer() ).getMovesWithScores().isEmpty();
        } else {
            return boardState.getLegalMoves().isEmpty();
        }

    }

    public List<Node> getBestNodes() {
        return bestNodes;
    }

    public void setInterrupted() {
        interrupted = true;
    }

    public boolean getInterrupted() { return interrupted; }

    public boolean getFinished() { return finished; }

    /**
     * Sort the nodes by most visited and add them to the bestNodes list.
     * @param children, the list of nodes to sort and find the best in.
     */
    private void addBestNode( List<Node> children ) {

        children.sort((Node o1, Node o2) -> Double.compare( o1.getVisits(), o2.getVisits() ) );
        bestNodes.add( children.get( children.size() - 1 ) );

    }

    /**
     *
     * @param currentBoard, the board to check the winner on
     * @return WinState, the actual WinState of the board
     */
    private WinState gameEnded ( Board currentBoard ) {
        WinState winner = currentBoard.checkWinner();
        return winner != null && !Objects.equals(winner.getWinner(), "None") ? winner : null;
    }

    /**
     * Has a player won the game yet?
     * @return Boolean, A player has won
     */
    private Player getWinner( Board currentBoard ) {
        WinState winner = currentBoard.checkWinner();
        return winner != null && winner.getWinner() != null && !Objects.equals(winner.getWinner(), "None") ? winner.getPlayer() : null;
    }

}
