package mcts;

import board.Board;
import board.WinState;
import game.Coordinate;
import player.Heuristic;
import player.Player;
import player.PositionAndScore;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Random;


/**
 * Created by danielford on 08/04/16.
 */
public class Node {

    private final Board board;
    private final Player player;
    private final Player opponent;
    private Node parent;

    private double visits = 0;
    private double wins = 0;
    private List<Coordinate> unplayedChildren;
    private List<Node> playedChildren;

    public Node(Board board, Node parent) {
        this.board = board.clone();
        this.parent = parent;

        this.board.setCurrentPlayer( board.getCurrentPlayerIndex() == 0 ? 1 : 0 );

        this.player = this.board.getCurrentPlayer();
        this.opponent = this.board.getPlayer( this.board.getCurrentPlayerIndex() == 0 ? 1 : 0 );
        unplayedChildren = new ArrayList<>();
        playedChildren = new ArrayList<>();

    }


    /**
     * Visits the node and if the current AI player has won it will increment it's wins
     * by 1. If it is a draw, the wins are incremented by 0.5 as even though it is not
     * a win, it is still better than losing.
     */
    public void visitNode( WinState gameEnd, Boolean isWinner ) {

        // If the player for this Node is the winner, increment wins!
        if ( isWinner ) {

            wins += 1;

        } else if ( gameEnd != null && Objects.equals(gameEnd.getWinner(), "Draw") ) {

            // If it's a draw, give it 'half a point' as it is better than losing
            wins += 0.5;

        }

        // This node has now been visited.
        visits++;

    }


    /**
     * If playedChildren is empty, that insinuates that there are no unplayed children,
     * this will then attempt to find the children of the current node.
     * These nodes will be randomly selected in the expansion phase and taken out of
     * the List 'unplayedChildren'.
     */
    public void findUnplayed() {

        // If it is not empty then don't expand as we have already expanded it
        if ( !playedChildren.isEmpty() ) { return; }

        unplayedChildren = normalNodes();


    }


    /**
     * Randomly selects a child from the 'unplayedChildren' list. Removes it and adds it to
     * the list 'playedChildren'.
     * @return Node, the randomly selected child.
     */
    public Node expand() {

        Random random = new Random();
        Board newBoard = board.clone();

        Coordinate move = unplayedChildren.get( random.nextInt( unplayedChildren.size() ));
        newBoard.createMove( move, opponent );

        Node node = new Node( newBoard, this );

        // Get this node read for the next iteration
        node.findUnplayed();

        playedChildren.add( node );
        unplayedChildren.remove( unplayedChildren.indexOf(move) );

        return node;

    }



    /**
     * Returns all possible moves for that node
     * @return List<Node>, all possible moves for a node
     */
    private List<Coordinate> normalNodes() {

        return board.getLegalMoves();

    }

    public double getVisits() { return visits; }

    public double getWins() { return wins; }

    public Board getBoard() { return board; }

    public Player getPlayer() { return player; }

    public Node getParent() {
        return parent;
    }

    public List<Coordinate> getUnplayedChildren() { return unplayedChildren; }

    public List<Node> getPlayedChildren() { return playedChildren; }

}
