package minimax;

import board.Board;
import game.Coordinate;
import player.PositionAndScore;

import java.awt.*;
import java.util.logging.Logger;


/**
 * Created by danielford on 04/04/16.
 */
public class IterativeDeepening {

    private long timeLimit;
    private final static Logger LOGGER = Logger.getLogger(IterativeDeepening.class.getName());
    private PositionAndScore bestSoFar;
    Board board;

    public IterativeDeepening(Board board, long timeLimit) {
        this.board = board;
        this.timeLimit = timeLimit;
    }

    public Coordinate makeMove( int startDepth ) {

        long start = System.currentTimeMillis();
        PositionAndScore best = null;

        // Create the deepening thread.
        DeepeningThread deepeningThread = new DeepeningThread(board, startDepth);
        deepeningThread.start();

        // While time limit not finished or deepening not finished
        while ( System.currentTimeMillis() < start + timeLimit || bestSoFar == null ) {

            // Get the best so far, might be null
            best = deepeningThread.getBestSoFar();

            // If null, sleep and try again in a second, else add as best move!
            if ( best == null ) {

                // Try again in 200 milliseconds as it's probably too soon to check for a bestMove
                try {
                    Thread.sleep(200);
                } catch (InterruptedException e) {
                    LOGGER.severe(String.valueOf(e));
                }

            } else {
                setBestSoFar( best );
            }

            // If finished, break!
            if ( deepeningThread.getFinished() ) {
                break;
            }

        }

        // Adds the best so far using deepeningThread (in case best was null)
        setBestSoFar( deepeningThread.getBestSoFar() );

        // Shut down the thread and minimax
        deepeningThread.setFinished();

        if ( bestSoFar == null ) {
            LOGGER.severe(" Computer has possibly exited with a null move. ");
            Toolkit.getDefaultToolkit().beep();
        } else {
            LOGGER.info("Computer exiting with best move");
        }

        // Return our best move
        return bestSoFar.getCoordinate();

    }

    // Try adding the proposed move as a new best.
    private void setBestSoFar( PositionAndScore best ) {

        if ( bestSoFar == null || best.getScore() > bestSoFar.getScore() ) {
            bestSoFar = best;
        }

    }

}
