package minimax;

import board.Board;
import board.WinState;
import game.Coordinate;
import player.Heuristic;
import player.Player;
import player.PositionAndScore;

import java.util.List;
import java.util.Objects;

/**
 * Created by danielford on 17/03/16.
 */
public class MiniMax {

    private Board board;
    private Player initialPlayer;
    private int initialPlayerIndex;
    private int opponentIndex;
    private Player currentPlayer;
    private int depthCompare;
    private PositionAndScore best;
    private boolean finished = false;

    public MiniMax(Board board, Player initialPlayer) {
        this.board = board;
        this.initialPlayer = initialPlayer;
    }

    public Coordinate makeMove(int startDepth) {

        PositionAndScore best = decideMove(startDepth);
        return best.getCoordinate();

    }

    private List<PositionAndScore> getMoves (Board state, Player currentPlayer) {
        return new Heuristic(state, currentPlayer).getMovesWithScores();
    }


    private void setupMinimax(int depth) {
        initialPlayerIndex = board.getCurrentPlayerIndex();
        opponentIndex = board.getCurrentPlayerIndex() == 0 ? 1 : 0;
        currentPlayer = initialPlayer;
        depthCompare = depth;
    }

    public PositionAndScore decideMove(int depth) {

        // Set up the important variables
        setupMinimax( depth );

        // Setup the initial moves.
        List<PositionAndScore> moves = getMoves(board, currentPlayer);
        best = null;
        double currentScore;

        // I.e. this is the 'default' starting move and should be played straight away.
        if ( moves.size() == 1 && moves.get(0).getScore() == 2000 ) {
            best = moves.get(0);
            return moves.get(0);
        }

        for ( PositionAndScore move: moves ) {


            // Make the AI player first again and take this move
            currentPlayer = board.setCurrentPlayer(initialPlayerIndex);
            board.createMove( move.getCoordinate(), currentPlayer );

            // AI player has taken move, so simulate the other player by minimising
            currentScore = minimiseMove(depth, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, move);

            // If  better than current best, replace best!
            if (best == null || currentScore > best.getScore()) {
                best = move;
                best.setScore( (int) currentScore );
            }

            // Reset the move as we're done with it.
            board.resetMove( move.getCoordinate() );

            if ( finished ) break;

        }

        // Set the current player to be the computer, ready for the next move
        board.setCurrentPlayer(initialPlayerIndex);

        return best;

    }

    /**
     * The Minimising stage of the MiniMax algorithm, plays from the opponents perspective where
     * they try maximising their score and therefore minimising the MiniMax players.
     * @param depth, the current depth
     * @param alpha, the current value of alpha from alpha-beta pruning
     * @param beta, the current value of beta from alpha-beta pruning
     * @param takenMove, the most recently taken move
     * @return Double, the result of the 'simulation'
     */
    public double minimiseMove( int depth, double alpha, double beta, PositionAndScore takenMove ) {

        double result;

        // Check the current score
        WinState winner = board.checkWinner();

        if ( depth == 0 || isGameOver( winner ) || finished ) {

            // Depth has been reached or game is over so score!
            result = scoreState( isGameOver( winner ), winner, depth, takenMove );

        } else {

            // Get the possible moves (Scored by the heuristic for the opponent)
            result = Double.POSITIVE_INFINITY;
            List<PositionAndScore> moves = getMoves(board, board.getPlayer( opponentIndex ));

            for ( PositionAndScore newMove : moves ) {


                // Create the move
                currentPlayer = board.setCurrentPlayer( opponentIndex );
                board.createMove( newMove.getCoordinate(), currentPlayer );

                // Score the move!
                double score = maximiseMove( depth - 1, alpha, beta, newMove );

                // We're done with it, remove it from the board.
                board.resetMove( newMove.getCoordinate() );

                // Minimise the move score
                result = Math.min(result, score);
                beta = Math.min(beta, score);

                // Out of bounds so break
                if ( beta <= alpha ) {
                    break;
                }

            }

        }

        return result;

    }

    /**
     * The Maximising stage of the MiniMax algorithm, plays from the MiniMax players perspective where
     * they try maximising their score and therefore minimising the opponents players.
     * @param depth, the current depth
     * @param alpha, the current value of alpha from alpha-beta pruning
     * @param beta, the current value of beta from alpha-beta pruning
     * @param takenMove, the most recently taken move
     * @return Double, the result of the 'simulation'
     */
    public double maximiseMove( int depth, double alpha, double beta, PositionAndScore takenMove ) {

        double result;

        // Check the current score
        WinState winner = board.checkWinner();

        if ( depth == 0 || isGameOver( winner ) || finished ) {

            // Depth has been reached or game is over so score!
            result = scoreState(isGameOver( winner ), winner, depth, takenMove );

        } else {

            // Get the possible moves (Scored by the heuristic for the current player)
            result = Double.NEGATIVE_INFINITY;
            List<PositionAndScore> moves = getMoves(board, board.getPlayer(initialPlayerIndex) );

            for (PositionAndScore newMove : moves ) {

                // Create the move
                currentPlayer = board.setCurrentPlayer( initialPlayerIndex );
                board.createMove( newMove.getCoordinate(), currentPlayer );

                // Score the move
                double score = minimiseMove( depth - 1, alpha, beta, newMove );

                // Remove the move, we're done with it.
                board.resetMove( newMove.getCoordinate() );

                // Maximise the move score
                result = Math.max(result, score);
                alpha = Math.max(alpha, score);

                // Out of bounds so break
                if ( beta <= alpha ) {
                    break;
                }

            }

        }

        return result;
    }


    /**
     * Scores the state using a base value of 100, the heuristic value and the current depth. This creates
     * a score which acts as a sort of confidence value for the algorithm to decide how good a move is. I.e.
     * if the algorithm has a move greater than 100, they know that they have one for sure.
     * @param gameOver, the current state of the board.
     * @param winner, The player who won
     * @param depth, The depth traversed, used for comparison against the initial specified dpeth
     * @param takenMove, the move taken to get to this state, used for heuristic scoring
     * @return Double, the heuristically evaluated score for this move.
     */
    private double scoreState(Boolean gameOver, WinState winner, int depth, PositionAndScore takenMove) {

        double result;

        // If draw, score is 0. Else get the heuristic value of the move
        result = gameOver && winner.getWinner().equals("Draw") ? 0 : takenMove.getScore();

        // Find the depth we've travelled, used as  a sort of 'Usefulness' heuristic
        depth = depthCompare - depth;

        // If it is not draw, somebody has one so score the minimiser / maximiser accordingly + a bonus of 100
        if ( gameOver && !winner.getWinner().equals("Draw")) {
            result = winner.getPlayer().equals( initialPlayer ) ? (result + 100) - depth : (result + depth) - 100;
        }

        return result;

    }

    private boolean isGameOver( WinState winner ) { return winner != null && winner.getWinner() != null && !Objects.equals(winner.getWinner(), "None"); }

    private boolean isInitialPlayer(Player player) {
        return player.equals(initialPlayer);
    }

    // Return the best move so found, useful incase there isn't enough time to finish e.g. in IterativeDeepening
    public PositionAndScore getBestSoFar() { return best; }

    // Sets finished to true to indicate minimax should stop
    public void setFinished() {
        finished = true;
    }


}
