package player;

import board.Board;
import game.Coordinate;
import game.SetupGame;
import javafx.application.Platform;
import javafx.scene.shape.Circle;
import mcts.MonteCarlo;
import minimax.IterativeDeepening;
import minimax.MiniMax;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Created by danielford on 02/03/16.
 */
public class Computer extends Player {

    String playerType;
    private Map<String, String> config = new HashMap<>();
    private final static Logger LOGGER = Logger.getLogger(Computer.class.getName());

    public Computer(int number, String color, String algorithm) {
        super(number, color);
        playerType = algorithm;
        createConfig();
    }

    public void takeMove( SetupGame state ) {

        // Clone the board and start the counter. The counter used to see how long the player has taken to make a move.
        Board clonedBoard = state.getBoard().clone();
        long startTime = System.currentTimeMillis();

        LOGGER.log(Level.INFO, "Computer player is thinking about a move...");

        // Retrieve the move as calculated by the AI player
        Coordinate move = makeMove( clonedBoard );

        // Create the move on the actual board's data layer
        state.getBoard().createMove(move, this);

        // Create the counter on the UI and progress to the next event
        Platform.runLater(() -> {
            createTile( state, move );
            state.nextEvent();
        });

        LOGGER.log(Level.INFO, "Computer player "+ getPlayerIdentity() +",  finished in " + (System.currentTimeMillis() - startTime) + "ms.");

    }

    // The default config that the algorithms will use
    private void createConfig() {
        config.put("maxDepth", "3");
        config.put("maxTime", "15000");
        config.put("maxMoves", "20000");
    }

    // Override the default config with user defined alternatives.
    public void overrideConfig(Map<String, String> newConfig) { newConfig.forEach(config::put); }

    /**
     * Create the move on the UI representation of the board.
     * @param state, the UI to place the move on
     * @param move, the move to place on the UI
     */
    private void createTile( SetupGame state, Coordinate move ) {
        Circle circle = state.getGame().getTile(move).getCircle();
        circle.getStyleClass().remove("empty");
        circle.getStyleClass().add("taken");
        circle.setFill(getPlayerColor());
    }


    /**
     * Returns a move made by the specified algorithm.
     * @param board, the cloned board to simulate on
     * @return Coordinate, the move to make on the board
     */
    private Coordinate makeMove( Board board ) {

        switch (playerType) {
            case "MiniMax":

                return new MiniMax(board, this).makeMove(Integer.parseInt(config.get("maxDepth")));

            case "MonteCarlo-Heavy":
            case "Monte Carlo":

                // This MonteCarlo will play well due to using heavy playouts through use of the Heuristic
                return new MonteCarlo(board, Integer.parseInt(config.get("maxTime")), Integer.parseInt(config.get("maxMoves")), true).makeMove();

            case "MonteCarlo-Light":

                // This one will play incredibly badly in Gomoku due to the large search space and using random playouts
                return new MonteCarlo(board, Integer.parseInt(config.get("maxTime")), Integer.parseInt(config.get("maxMoves")), false).makeMove();

            case "Heuristic":

                return new Heuristic(board, board.getCurrentPlayer()).makeMove();

            default:

                return new IterativeDeepening(board, Integer.parseInt(config.get("maxTime"))).makeMove( Integer.parseInt(config.get("maxDepth")) );

        }

    }

    /**
     * For test purposes, lets the player play with just the data layer
     */
    public void testCreateMove(Board board) {

        Board clonedBoard = board.clone();
        Coordinate move = makeMove( clonedBoard );
        board.createMove(move, this);

    }

}
