package player;

import board.Board;
import board.WinState;
import game.Coordinate;
import javafx.geometry.Pos;

import javax.swing.text.Position;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by danielford on 02/03/16.
 */
public class Heuristic {

    private Board board;
    private Board tempBoard;
    private int boardSize;
    private Player currentPlayer;
    private List<PositionAndScore> scoredMoves;


    public Heuristic(Board board, Player currentPlayer) {
        this.board = board;
        this.currentPlayer = currentPlayer;
        boardSize = board.getDimensions();

    }

    public Coordinate makeMove() {
        decideMove();
        return getBest().getCoordinate();
    }


    public List<PositionAndScore> getMovesWithScores() {

        scoredMoves = new ArrayList<>();
        tempBoard = board;

        for ( int i = 0; i < boardSize; i++ ) {

            for ( int j = 0; j < boardSize; j++ ) {

                Coordinate position = new Coordinate(i, j);
                Player boardPosition = tempBoard.getPosition(position);

                if ( boardPosition  != null ) {

                    findRow(position);
                    findColumn(position);
                    findLeftDiagonal(position);
                    findRightDiagonal(position);

                }

            }

        }


        // If there are no moves, this insinuates nobody has played yet. So just place in the middle.
        // TODO: This is probably pretty poor, replace with an opening book of best starting moves?
        if ( scoredMoves.size() <= 0 ) {
            scoredMoves.add(new PositionAndScore( centralMove(), 2000, true));
        } else {
            scoredMoves.sort((PositionAndScore a, PositionAndScore b) -> Double.compare(b.getScore(), a.getScore()));
        }


        return scoredMoves;

    }

    private Coordinate centralMove() {
        return new Coordinate( boardSize / 2, boardSize / 2 );
    }


    /**
     * Similar to the method used to check for the winner in the Board class. This algorithm starts at some start
     * point and then counts the counters across until some end point. It then changes direction and counts the
     * counters the other way. It then heuristically evaluates the spaces at either end of the counters.
     * @param checkRow, whether or not the algorithm should check the rows for counters
     * @param checkCol, whether or not the algorithm should check columns for counters
     * @param checkDiag, whether or not the algorithm should check right diagonals for counters
     * @param lastMove, the current move given to the algorithm to start its search
     */
    private void findMoves(Boolean checkRow, Boolean checkCol, Boolean checkDiag, Coordinate lastMove) {

        if ( lastMove == null ) return;

        boolean notFinished = true;
        Coordinate pointer = lastMove;
        List<Coordinate> possibleMoves = new ArrayList<>();
        int newCol = pointer.column;
        int newRow = pointer.row;
        int currentScore = 0;
        int winningScore = 5;

        while ( currentScore != winningScore ) {

            // We are moving the column value i.e. checkCol
            if ( checkCol && !checkRow ) {
                newCol = notFinished ? pointer.column - 1 : pointer.column + 1;
            } else if ( checkRow && !checkCol) {
                // We are moving the row value i.e.  checkRow
                newRow = notFinished ? pointer.row - 1 : pointer.row + 1;
            } else if ( checkDiag ) {
                newCol = notFinished ? pointer.column - 1 : pointer.column + 1;
                newRow = notFinished ? pointer.row - 1 : pointer.row + 1;
            } else {
                newCol = notFinished ? pointer.column + 1 : pointer.column - 1;
                newRow = notFinished ? pointer.row - 1 : pointer.row + 1;
            }

            // If there is a player in the position, and they are the current player, increase counter
            if ( tempBoard.getPosition(pointer) != null &&
                    tempBoard.getPosition(pointer).getPlayerIdentity() == tempBoard.getPosition(lastMove).getPlayer().getPlayerIdentity() ) {

                pointer = new Coordinate(newCol, newRow);

                currentScore++;


            } else if ( notFinished ) {

                if ( (pointer.row >= 0 && pointer.row <= tempBoard.getDimensions() - 1) &&
                        (pointer.column >= 0 && pointer.column <= tempBoard.getDimensions() -1) ) {
                    possibleMoves.add(pointer);
                }

                notFinished = false;
                int resetCol = lastMove.column;
                int resetRow = lastMove.row;

                if ( checkCol && !checkRow ) {
                    resetCol = resetCol + 1;
                } else if ( checkRow && !checkCol ) {
                    resetRow = resetRow + 1;
                } else if ( checkDiag ) {
                    resetCol = resetCol + 1;
                    resetRow = resetRow + 1;
                } else {
                    resetCol = resetCol - 1;
                    resetRow = resetRow + 1;
                }

                pointer = new Coordinate(resetCol, resetRow);

            } else {

                if ( (pointer.row >= 0 && pointer.row <= tempBoard.getDimensions() - 1) &&
                        (pointer.column >= 0 && pointer.column <= tempBoard.getDimensions() -1) ) {
                    possibleMoves.add(pointer);
                }

                break;
            }

        }

        // Passes the possible moves, the score for those moves, and whether or not this is for the current player
        scoreMoves(possibleMoves, currentScore, board.getPosition(lastMove) != null && board.getPosition(lastMove).equals( currentPlayer ));
    }

    /**
     * Scores the moves negatively or positively depening on if it is the player or their opponent. If the move already
     * in the list then it is evaluated further.
     * @param moves, the move to score
     * @param score, the score found so far for the move
     * @param isPlayer, if the move is for the current player
     */
    private void scoreMoves(List<Coordinate> moves, int score, boolean isPlayer) {

        for(Coordinate move : moves) {

            boolean added = false;

            if ( tempBoard.getPosition(move) == null ) {

                for ( PositionAndScore scored : scoredMoves ) {

                    if ( move.compareMove( scored.getCoordinate() )) {

                        PositionAndScore scoredMove = scoredMoves.get(scoredMoves.indexOf(scored));
                        scoredMove.setScore( scoredMove.getScore() + (isPlayer ? score : -score) );
                        added = true;
                        break;

                    }

                }

                if ( !added  ) {
                    scoredMoves.add(new PositionAndScore(move, isPlayer ? score : -score, isPlayer));
                }

            }

        }

    }

    private void findLeftDiagonal(Coordinate position) {
        findMoves(true, true, false, position);
    }

    private void findRightDiagonal(Coordinate position) { findMoves(true, true, true, position); }

    private void findColumn(Coordinate position) {
        findMoves(false, true, false, position);
    }

    private void findRow(Coordinate position) {
        findMoves(true, false, false, position);
    }

    private void decideMove() {

        // Calls findRow, findColumn and findDiagonal to find possible moves
        getMovesWithScores();

    }

    /**
     * When playing with just the heuristic, this is a greedy method of getting the absolute
     * current best move for the heuristic to take.
     * @return PositionAndScore, the scored move for the heuristic to take
     */
    private PositionAndScore getBest() {

        int score;
        int max = Integer.MIN_VALUE;
        PositionAndScore bestMove = null;

        for( PositionAndScore move : scoredMoves ) {

            score = move.getScore();

            if(max < Math.abs(score) ) {
                max = Math.abs(score);
                bestMove = move;
            }

        }

        return bestMove;
    }

}
