import board.Board;
import game.Coordinate;
import mcts.MonteCarlo;
import org.junit.Assert;
import org.junit.Test;
import player.Computer;
import player.Human;
import player.Player;

/**
 * Created by danielford
 */
public class MCTSTests {


    Board board;

    private void setupBoard() {
        board = new Board(19, 5);
        board.createPlayer( new Computer(1, "BLACK", "MonteCarlo") );
        board.createPlayer( new Human(2, "WHITE") );
        board.setCurrentPlayer(0);
    }


    /**
     * The Board is set up with 4 AI player counters in a row.
     * The move that the monte carlo algorithm should pick would be the one
     * to complete this row. e.g. 5,9 or 10,9
     */
    @Test
    public void sureWinRow() {

        setupBoard();
        Player computer = board.getCurrentPlayer();
        Player human = board.getPlayer(1);

        board.createMove(new Coordinate(9, 9), computer);
        board.createMove(new Coordinate(8, 9), computer);
        board.createMove(new Coordinate(7, 9), computer);
        board.createMove(new Coordinate(6, 9), computer);

        board.createMove(new Coordinate(9, 15), human);
        board.createMove(new Coordinate(8, 15), human);
        board.createMove(new Coordinate(7, 15), human);
        board.createMove(new Coordinate(6, 15), human);

        Coordinate move = new MonteCarlo(board, 10000, 2000, true).makeMove();

        System.out.println("Picked: " + "Col: " + move.column + " Row:" + move.row);

        board.createMove(move, computer);

        Assert.assertTrue(board.checkWinner().getPlayer() != null && board.checkWinner().getPlayer().equals(computer));

    }

    /**
     * The Board is set up with 4 AI player counters in a row with an opponent in the same condition.
     * The move that the monte carlo algorithm should pick would be the one that leads it to win
     * instead of blocking the opponent.
     */
    @Test
    public void sureWinWithOpponentDiag() {

        setupBoard();
        Player player = board.getCurrentPlayer();
        Player human = board.getPlayer(1);

        // Computer moves
        board.createMove(new Coordinate(8, 8), player);
        board.createMove(new Coordinate(9, 9), player);
        board.createMove(new Coordinate(10, 10), player);
        board.createMove(new Coordinate(11, 11), player);

        // Create moves for human
        board.createMove(new Coordinate(5, 5), human);
        board.createMove(new Coordinate(4, 4), human);
        board.createMove(new Coordinate(3, 3), human);
        board.createMove(new Coordinate(2, 2), human);


        Coordinate move = new MonteCarlo(board, 10000, 2000, true).makeMove();

        board.createMove(move, player);

        Assert.assertTrue(board.checkWinner().getPlayer() != null && board.checkWinner().getPlayer().equals(player));

    }


    /**
     * The Board is set up with the Computer player nearly having 4 in a row.
     * The move that the monte carlo algorithm should pick would be the one that breaks the
     * four in the row from happening i.e. column 2 and row 2.
     */
    @Test
    public void defensiveMove() {

        setupBoard();
        Player computer = board.getCurrentPlayer();
        Player human = board.getPlayer(1);

        // Round 1
        board.createMove(new Coordinate(5, 5), human);
        board.createMove(new Coordinate(2, 3), computer);

        // Round 2
        board.createMove(new Coordinate(4, 4), human);
        board.createMove(new Coordinate(2, 4), computer);

        // Round 3
        board.createMove(new Coordinate(1, 1), human);
        board.createMove(new Coordinate(2, 1), computer);

        // Round 4
        board.createMove(new Coordinate(3, 3), human);
        Coordinate move = new MonteCarlo(board, 15000, 2000, true).makeMove();


        if ( move.column != 2 && move.row !=2 ) {
            System.out.println("STOP");
        }

        Assert.assertTrue( move.column == 2 && move.row == 2 );

    }

//    @Test
//    public void runLoads() {
//
//        for ( int i = 0; i < 100; i++ ) {
//            sureLossWithOpponent();
//        }
//
//    }


}
